/************************************************************************************************
 * Copyright  (c) 2011, 2012 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ***********************************************************************************************/

import Vector::*;

typedef 64 PAYLOAD_WIDTH;
typedef 46 PTR_WIDTH;
typedef 6 RF_WIDTH;
typedef 6 PTR_FIELD_WIDTH;

//defining various data types and structures for the processor
typedef Bit#(PAYLOAD_WIDTH) Payload; //payload size
typedef UInt#(PAYLOAD_WIDTH) TimerData;
typedef Bit#(7) OpCode;

//instruction immediate width
typedef Bit#(10) Imm;

typedef Bit#(28) TAG_A;

typedef Bit#(RF_WIDTH) RF_A;

//fatptr related
typedef Bit#(PTR_FIELD_WIDTH) I;
typedef Bit#(PTR_FIELD_WIDTH) B;
typedef Bit#(PTR_FIELD_WIDTH) M;
typedef Bit#(PTR_FIELD_WIDTH) L;
typedef Bit#(PTR_WIDTH) A;

typedef Int#(PTR_WIDTH) Dunder;
typedef Int#(PTR_WIDTH) Dover;

typedef struct {
        Bit#(1) dummy;
        B b;
        M m;
        A a;
} Tag deriving (Bits, Eq, Bounded);

typedef struct{
        AtomicGroup ag;
        Tag tag;
} MetaData deriving (Bits, Eq, Bounded);

//Generic Form of data
typedef struct{
        MetaData meta;
        Payload payload;
} Atom deriving (Bits, Eq, Bounded);

typedef struct{
        Atom a0;
        Atom a1;
        Atom a2;
        Atom a3;
} CacheLine deriving (Bits, Eq, Bounded);

//Timer Registers (qt, clk, slot); Unsigned Int for values
typedef struct{
        MetaData meta;
        TimerData value;
} TimerReg deriving (Bits, Eq, Bounded);

typedef struct{
        MetaData meta;
        B b;
        I i;
        M m;
        A a;
} FatPointer deriving (Bits, Eq, Bounded);

typedef struct{
        Atom atom;
        Dunder dunder;
        Dover dover;
} RegAtom deriving (Bits, Eq, Bounded);

typedef Bit#(4) LVTOut;
typedef Vector#(2, RF_A) WriteIndices;
typedef Vector#(4, RF_A) ReadIndices;
typedef struct{
	RegAtom val0;
	RegAtom val1;
	RegAtom val2;
	RegAtom val3;
} RFOut deriving (Bits, Eq, Bounded);

typedef 32 NumGPR;
//32 deep register file indices
typedef enum {R0, R1, R2, R3, R4, R5, R6, R7, R8,
              R9, R10, R11, R12, R13, R14, R15, R16,
              R17, R18, R19, R20, R21, R22, R23, R24,
              R25, R26, R27, R28, R29 ,R30, R31
} GPR deriving (Bits, Eq, Bounded);

//instruction opcodes	= 80 (06/29)
typedef enum {ADD, SUB, LCFP, CPMR, CPRM,
              NOP, HALT, JMP, BEQ, GCALL,
              GRTN, RUNT, YIELD, GTIME, REMT,
              RDTIME, STRB, STRY, STWB, STWY,
              AND, OR, SHR, SHL, CPRR,
              MVRR, FRAMPTR, OFFP, OFFLP, TMUL,
              TMUU, RETAG, INTAG, MAKEP, NEWT,
              INP, SETA, REGRP, MVRM, MVMR,
              MUL, RESUMET, TEST, GATE, FJMP,
              BNE, NEWG, BNEG, TOTAG, RFLCT,
              GFWRITE, FPHASH, RFLCTP, RETAGPC, SETTAG,
              CALL, MAKET, INGRP, TAGOF, GJMP,
              BPOS, TMURC, INA, CPMT, CPIO,
              CPAR, CPHR, OFFTP, TRANSBEGIN, TRANSEND,
              GACALL, GAJMP, TESTGRP, YIELD2, CPMP,
              CPPM, ENDU, RDFHTP, SETFHTP, CWBM,
	      BCALL, BGCALL, BGACALL, BRTN
} OPCODE deriving (Bits, Eq, Bounded); //TODO: need insn for flushing

//atomic groups
typedef enum {FR_POINTER, CONST_POINTER, LIN_FR_POINTER, 
              INSTR_POINTER, GATE_POINTER, STREAM_READ_POINTER, 
              STREAM_WRITE_POINTER, THREAD_POINTER, FWD_POINTER, 
              DOUBLE, INTEGER, INSN, 
              AUTHORITY, PRINCIPAL, UNINITIALIZED, 
              ERROR, TAG, EMPTY
} AtomicGroup deriving (Bits, Eq, Bounded);

//thread states
typedef enum {RUNNING, RUNNABLE, UNINITIALIZED,
              INITIALIZED, YIELD_TIMER, VOL_YIELD,
              FAULTED, HALTED
} ThreadState deriving (Bits, Eq, Bounded);

//error codes
typedef enum {NO_ERROR, TMU_FAULT, STREAM_BLOCKING,
              ARITH_ERROR, FRAME_BOUNDS_ERROR, AG_ERROR,
              FP_ALIGN_ERROR, GS_BOUNDS_ERROR, TRANS_ERROR,
              MANGLED_THREAD_FRAME, TIMER_INTERRUPT
} ErrorCode deriving (Bits, Eq, Bounded);

//decoder interprets insn that have imm and of what width
typedef struct{
        MetaData meta;
        Bit#(32) dummy;
        OpCode opcode;
        GPR dest;
        GPR src1;
        GPR src2;
        Imm imm;
} Instruction deriving (Bits, Eq, Bounded);

//output of a decoder
typedef struct{
        OPCODE op;
        GPR dest;
        GPR src1;
        GPR src2;
        Bit#(25) imm;
} DecodedSignal deriving (Bits, Eq, Bounded);

//thread status structure
typedef struct{
        ThreadState thread_state;
        ErrorCode error_code;
} ThreadStatus deriving (Bits, Eq);

typedef struct {
        TAG_A tptag;
        TAG_A citag;
        Bit#(8) opgrp;
        TAG_A pctag;
        TAG_A auth;
        TAG_A op1tag;
        TAG_A op2tag;
        TAG_A op3tag;
        TAG_A mrtag;
} TMU_M deriving (Bits, Eq, Bounded);

typedef struct {
        TAG_A op3res;
        TAG_A op2res;
        TAG_A op1res;
        TAG_A mwres;
        TAG_A pcres;
        TAG_A tpres;
        TAG_A auth;
} TMU_R deriving (Bits, Eq, Bounded);

typedef struct {
        TMU_M match_vector;
        TMU_R res_vector;
} TMU_MR deriving (Bits, Eq, Bounded);

//Output types of different functional units
typedef Tuple2# (Bool, AtomicGroup) AGUOut;
typedef Tuple2# (Bool, FatPointer) FatPUOut;
typedef Tuple2# (Bool, Payload) ALUOut;
typedef Tuple2# (Bool, TimerData) TimerOut;
typedef Tuple2# (Bool, FatPointer) PCUOut;

//all the private state indices for vectors
typedef 0 PC;
typedef 1 AUTH;
typedef 2 TS;
typedef 3 GSP;
typedef 4 CI;
typedef 5 OP1;
typedef 6 OP2;
typedef 7 OP3;
typedef 8 MR;
typedef 9 TMUR;
typedef 10 FPC;
typedef 11 FA;
typedef 12 H1;
typedef 13 H2;
typedef 16 TP;